from typing import Optional, Type

import polars as pl

from ldnode.calculator.schemas import (
    SecurityPoolType,
    SecurityPoolCustomizeType,
    SecurityPool,
    ExcludeOperator,
    BuiltinScope,
)


def _gen_customized_pool_expr(security_pool: SecurityPool) -> Optional[pl.Expr]:
    """ 生成自定义标的池表达式 """
    match security_pool.get('pool_customize_type'):
        case SecurityPoolCustomizeType.FIXED:
            if securities := security_pool.get('pool_securities'):
                return pl.col('code').is_in(securities)
        case SecurityPoolCustomizeType.DYNAMIC_BY_FACTOR:
            if filter_factors := security_pool.get('pool_filter_factors'):
                filter_exprs = [
                    ExcludeOperator(filter_factor['operator']).compare(filter_factor['factor'], filter_factor['value'])
                    for filter_factor in filter_factors
                    if filter_factor['operator'] in ExcludeOperator
                ]
                return pl.Expr.and_(*filter_exprs)
            

def _gen_pool_expr(security_pool: SecurityPool, BulitinScopes: Type[BuiltinScope]) -> Optional[pl.Expr]:
    """ 生成标的池表达式 """
    match security_pool['pool_type']:
        case SecurityPoolType.BUILTIN:
            if builtin_scope := security_pool.get('pool_builtin_scope'):
                if builtin_scope in BulitinScopes:
                    return BulitinScopes(builtin_scope).process()
        case SecurityPoolType.CUSTOMIZED:
            return _gen_customized_pool_expr(security_pool)


def _gen_pool_exprs(security_pools: list[SecurityPool], BulitinScopes: Type[BuiltinScope]) -> list[pl.Expr]:
    """ 生成标的池表达式列表 """
    return list(
        filter(
            lambda x: x is not None,
            [_gen_pool_expr(security_pool, BulitinScopes) for security_pool in security_pools], 
        )
    )


def planning_security_pools(
    frame: pl.LazyFrame | pl.DataFrame,
    BuiltinScopes: Type[BuiltinScope],
    security_pools: Optional[list[SecurityPool]],
) -> pl.LazyFrame:
    """ 
        规划股票池计算 
        前置依赖: 规划百分比因子计算
    """
    lf = frame.lazy() if isinstance(frame, pl.DataFrame) else frame
    # 兜底：确保 code 列为 String 类型（cb_data.pq 中 code 为 Categorical，
    # BuiltinScope 的 str.starts_with 等操作要求 String）
    if 'code' in lf.collect_schema().names():
        lf = lf.with_columns(pl.col('code').cast(pl.String))
    # 若没有标的池，则直接返回
    if not security_pools:
        return lf.with_columns(pl.lit(True).alias('_isin_pool'))
    # 标的池表达式
    if pool_exprs := _gen_pool_exprs(security_pools, BuiltinScopes):
        pool_expr = pl.Expr.or_(*pool_exprs).fill_null(False)
        # 曾入池的标的保留全部数据
        ever_in_pool_codes = (
            lf.filter(pool_expr)
            .select('code')
            .unique()
            .collect()
            .get_column('code')
        )
        # 只保留曾经入选标的池的标的
        return lf.filter(
            pl.col('code').is_in(ever_in_pool_codes)
        ).with_columns(
            (pool_expr).alias('_isin_pool')
        )
    else:
        return lf.with_columns(pl.lit(True).alias('_isin_pool'))
    


CB_SCREEN_RESULT_FIXED_COLUMNS: list[str] = [
    'code', 'name', 'time', 'rank', '_weight', 'selected', 'close', 'pct_chg', 'amount', 'limit',
]


STK_SCREEN_RESULT_FIXED_COLUMNS: list[str] = [
    'code', 'name', 'time', 'rank', '_weight', 'selected', 'close', 'pct_chg', 'amount', 'limit',
]


ETF_SCREEN_RESULT_FIXED_COLUMNS: list[str] = [
    'code', 'name', 'time', 'rank', '_weight', 'selected', 'close', 'pct_chg', 'amount', 'limit',
]
""" ETF截面结果固定列 """



import abc
import json
from enum import IntEnum
from dataclasses import dataclass
from pathlib import Path
from typing import TypedDict, Optional, Callable
from threading import Lock

import polars as pl

from .enumerations import SyncStatus, LoadTrigger


@dataclass
class SyncResult:
    filename: str
    status: SyncStatus
    latest_tag_or_error: str


class DataSource(abc.ABC):
    """ 数据源 """
    @abc.abstractmethod
    def load(self, key: str) -> list[SyncResult]:
        """ 加载数据 """
        ...


class FileType(IntEnum):
    """ 文件类型 """
    PARQUET = 0 # parquet文件
    DELTALAKE = 1 # deltalake文件

        
@dataclass
class File:
    """ 文件 """
    key: str
    file_type: FileType = FileType.PARQUET

    def get_local_path(self, root_path: Path | None = None) -> Path:
        """ 获取本地路径 """
        if not root_path:
            root_path = Path("./data")
        return root_path / self.key

    def output_lazyframe(self, root_path: Path | None = None) -> pl.LazyFrame:
        """ 输出懒惰框架（始终读本地文件；S3 行情由 LoaderWorker 后台预拉取写入本地）"""
        while True:
            try:
                match self.file_type:
                    case FileType.PARQUET:
                        return pl.scan_parquet(self.get_local_path(root_path))
                    case FileType.DELTALAKE:
                        return pl.scan_delta(self.get_local_path(root_path))
                    case _:
                        raise ValueError(f"不支持的文件类型: {self.file_type}")
            except IOError:
                continue

    def output_eager_lazyframe(self, root_path: Path | None = None) -> pl.LazyFrame:
        """ 即时读取后转 LazyFrame（适用于频繁更新的小文件，避免 scan→collect 竞态）
        catch Exception 而非仅 IOError：parquet 竞态抛出 ComputeError，非 IOError """
        while True:
            try:
                return pl.read_parquet(self.get_local_path(root_path)).lazy()
            except Exception:
                continue
            

@dataclass
class LoadRule:
    """ 加载规则 """
    file: File
    tigger: LoadTrigger
    datasource: DataSource
    interval: int | None = None
    is_stop: Callable[[File], bool] | None = None


class Metadata(TypedDict):
    etags: dict[str, str]


class SyncMetadata:
    """
        同步元数据
    """
    METADATA_FILENAME = ".sync_metadata.json"
    
    def __init__(self, local_data_dir: Path) -> None:
        self.lock = Lock()
        self.metadata_file = local_data_dir / self.METADATA_FILENAME
        self.metadata = self._load_metadata()
    
    def _load_metadata(self) -> Metadata:
        """
            加载元数据
            @return: 元数据
        """
        with self.lock:
            if self.metadata_file.exists():
                with open(self.metadata_file, 'r', encoding='utf-8') as f:
                    return json.load(f)
            else:
                return {"etags": {}}
        
    def _save_metadata(self) -> None:
        """
            保存元数据
        """
        with open(self.metadata_file, 'w', encoding='utf-8') as f:
            json.dump(self.metadata, f, ensure_ascii=False, indent=2)
        
    def get_file_etag(self, filename: str) -> Optional[str]:
        """
            获取文件ETAG
            @param filename: 文件名
            @return: 文件ETAG
        """
        return self.metadata.get('etags', {}).get(filename, None)
    
    def set_file_etag(self, filename: str, etag: str) -> None:
        """
            设置文件ETAG
            @param filename: 文件名
            @param etag: 文件ETAG
        """
        with self.lock:
            self.metadata['etags'][filename] = etag
            self._save_metadata()
        
    def delete_file_etag(self, filename: str) -> None:
        """
            删除文件ETAG
            @param filename: 文件名
        """
        with self.lock:
            self.metadata['etags'].pop(filename, None)
            self._save_metadata()



"""
Issue #8 & #9 单元测试

Issue #8: 滚动 IC 均值算法修复（_calc_ic_series_data 中的 rolling 逻辑）
Issue #9: jaccard_overlap_matrix 新增 Overlap 矩阵字段
"""

import pytest
import polars as pl
from datetime import date

from ldnode.calculator.operators.correlation import jaccard_overlap_matrix
from ldnode.calculator.schemas.enumerations import SecurityType


# ---------------------------------------------------------------------------
# 辅助函数
# ---------------------------------------------------------------------------

def compute_rolling(ic_list: list, rolling_period: int) -> list:
    """复制被测算法逻辑（与 _calc_ic_series_data 中 rolling 段完全一致）"""
    rolling = []
    n = len(ic_list)
    for i in range(n):
        window = ic_list[max(0, i - rolling_period + 1): i + 1]
        if len(window) < rolling_period:
            rolling.append(None)
        else:
            total = sum(v if v is not None else 0.0 for v in window)
            rolling.append(total / rolling_period)
    return rolling


def make_res_df(dates_holdings: list[tuple[str, list[str] | None]]) -> pl.DataFrame:
    """构造包含 trade_date 和 hold_securities 列的 res_df"""
    dates = [date.fromisoformat(d) for d, _ in dates_holdings]
    holdings = [h for _, h in dates_holdings]
    return pl.DataFrame({
        "trade_date": pl.Series(dates, dtype=pl.Date),
        "hold_securities": pl.Series(holdings, dtype=pl.List(pl.Utf8)),
    })


# ---------------------------------------------------------------------------
# Issue #8：滚动 IC 均值算法测试
# ---------------------------------------------------------------------------

class TestRollingICMean:
    """Issue #8：测试 _calc_ic_series_data 中的 rolling 均值计算逻辑"""

    def test_normal_sequence_no_none_period3(self):
        """正常序列无 None，period=3：前2个为 None，后3个为固定窗口均值"""
        ic_list = [0.1, 0.2, 0.3, 0.4, 0.5]
        result = compute_rolling(ic_list, rolling_period=3)

        assert result[0] is None
        assert result[1] is None
        assert result[2] == pytest.approx((0.1 + 0.2 + 0.3) / 3)
        assert result[3] == pytest.approx((0.2 + 0.3 + 0.4) / 3)
        assert result[4] == pytest.approx((0.3 + 0.4 + 0.5) / 3)

    def test_contains_none_period3(self):
        """含 None：[0.6, None, 0.4]，period=3 → 前2个 None，第3个=(0.6+0+0.4)/3"""
        ic_list = [0.6, None, 0.4]
        result = compute_rolling(ic_list, rolling_period=3)

        assert result[0] is None
        assert result[1] is None
        assert result[2] == pytest.approx((0.6 + 0.0 + 0.4) / 3)

    def test_all_none_period3(self):
        """全 None：period=3 → 前2个 None，第3个=0.0"""
        ic_list = [None, None, None]
        result = compute_rolling(ic_list, rolling_period=3)

        assert result[0] is None
        assert result[1] is None
        assert result[2] == pytest.approx(0.0)

    def test_boundary_period1_with_none(self):
        """边界 period=1：[None, 0.5] → [0.0, 0.5]（None→0.0，非 None 原值）"""
        ic_list = [None, 0.5]
        result = compute_rolling(ic_list, rolling_period=1)

        assert result[0] == pytest.approx(0.0)
        assert result[1] == pytest.approx(0.5)

    def test_single_element_period1(self):
        """单元素 period=1：[0.3] → [0.3]"""
        ic_list = [0.3]
        result = compute_rolling(ic_list, rolling_period=1)

        assert len(result) == 1
        assert result[0] == pytest.approx(0.3)


# ---------------------------------------------------------------------------
# Issue #9：jaccard_overlap_matrix 新增 Overlap 字段测试
# ---------------------------------------------------------------------------

class TestJaccardOverlapMatrix:
    """Issue #9：测试 jaccard_overlap_matrix 的 jaccard 和 overlap 双矩阵输出"""

    def test_single_strategy_returns_diagonal_one(self):
        """单策略（n=1）→ labels=["A"], jaccard=[[1.0]], overlap=[[1.0]]"""
        res_df_map = {
            "A": make_res_df([("2024-01-01", ["000001", "000002"])])
        }
        security_type_map = {"A": SecurityType.STK}

        result = jaccard_overlap_matrix(res_df_map, security_type_map)

        assert result["labels"] == ["A"]
        assert result["jaccard"] == [[1.0]]
        assert result["overlap"] == [[1.0]]

    def test_empty_res_df_map_returns_empty(self):
        """空 res_df_map（n=0）→ labels=[], jaccard=[], overlap=[]"""
        result = jaccard_overlap_matrix({}, {})

        assert result["labels"] == []
        assert result["jaccard"] == []
        assert result["overlap"] == []

    def test_same_type_completely_disjoint(self):
        """同品种两策略，完全不重叠 → jaccard=0.0, overlap=0.0"""
        res_df_map = {
            "A": make_res_df([("2024-01-01", ["000001", "000002"])]),
            "B": make_res_df([("2024-01-01", ["000003", "000004"])]),
        }
        security_type_map = {"A": SecurityType.STK, "B": SecurityType.STK}

        result = jaccard_overlap_matrix(res_df_map, security_type_map)

        assert result["jaccard"][1][0] == pytest.approx(0.0)
        assert result["overlap"][1][0] == pytest.approx(0.0)

    def test_same_type_completely_overlapping(self):
        """同品种两策略，完全重叠 → jaccard=1.0, overlap=1.0"""
        res_df_map = {
            "A": make_res_df([("2024-01-01", ["000001", "000002"])]),
            "B": make_res_df([("2024-01-01", ["000001", "000002"])]),
        }
        security_type_map = {"A": SecurityType.STK, "B": SecurityType.STK}

        result = jaccard_overlap_matrix(res_df_map, security_type_map)

        assert result["jaccard"][1][0] == pytest.approx(1.0)
        assert result["overlap"][1][0] == pytest.approx(1.0)

    def test_same_type_partial_overlap(self):
        """同品种两策略 A={1,2,3}, B={2,3,4} → jaccard=2/4=0.5, overlap=2/3≈0.667"""
        res_df_map = {
            "A": make_res_df([("2024-01-01", ["s1", "s2", "s3"])]),
            "B": make_res_df([("2024-01-01", ["s2", "s3", "s4"])]),
        }
        security_type_map = {"A": SecurityType.STK, "B": SecurityType.STK}

        result = jaccard_overlap_matrix(res_df_map, security_type_map)

        # intersection=2, union={s1,s2,s3,s4}=4 → jaccard=2/4=0.5
        assert result["jaccard"][1][0] == pytest.approx(0.5)
        # intersection=2, min(3,3)=3 → overlap=2/3
        assert result["overlap"][1][0] == pytest.approx(2.0 / 3.0)

    def test_cross_security_type_returns_zero(self):
        """跨品种策略对 → jaccard=0.0, overlap=0.0"""
        res_df_map = {
            "stk_a": make_res_df([("2024-01-01", ["000001", "000002"])]),
            "cb_a": make_res_df([("2024-01-01", ["110001", "110002"])]),
        }
        security_type_map = {"stk_a": SecurityType.STK, "cb_a": SecurityType.CB}

        result = jaccard_overlap_matrix(res_df_map, security_type_map)

        assert result["jaccard"][1][0] == pytest.approx(0.0)
        assert result["overlap"][1][0] == pytest.approx(0.0)

    def test_hold_securities_none_does_not_raise(self):
        """hold_securities 为 null（Python None）→ 不抛 TypeError，intersection=0"""
        res_df_map = {
            "A": make_res_df([("2024-01-01", None)]),
            "B": make_res_df([("2024-01-01", ["000001"])]),
        }
        security_type_map = {"A": SecurityType.STK, "B": SecurityType.STK}

        # 不应抛出 TypeError
        result = jaccard_overlap_matrix(res_df_map, security_type_map)

        # None 视为空集，intersection=0，union={000001}=1 → jaccard=0.0
        assert result["jaccard"][1][0] == pytest.approx(0.0)
        # min(0, 1)=0 → overlap=0.0
        assert result["overlap"][1][0] == pytest.approx(0.0)
